@@ -1,13 +1,25 @@
 /*
+ * GTK VNC Widget
+ *  
  * Copyright (C) 2006  Anthony Liguori <anthony@codemonkey.ws>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser General Public License version 2 or
- * later as published by the Free Software Foundation.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.0 of the License, or (at your option) any later version.
  *
- *  GTK VNC Widget
+ * This library is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
  */
 
+#include <config.h>
+
 #include "vncdisplay.h"
 #include "coroutine.h"
 #include "gvnc.h"
@@ -24,13 +36,17 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#ifdef HAVE_PWD_H
 #include <pwd.h>
-
-#if WITH_GTKGLEXT
-#include <gtk/gtkgl.h>
-#include <GL/gl.h>
 #endif
 
+#ifdef HAVE_WINSOCK2_H
+#include <winsock2.h>
+#endif
+
+static void winsock_startup (void);
+static void winsock_cleanup (void);
+
 #define VNC_DISPLAY_GET_PRIVATE(obj) \
       (G_TYPE_INSTANCE_GET_PRIVATE((obj), VNC_TYPE_DISPLAY, VncDisplayPrivate))
 
@@ -41,22 +57,10 @@ struct _VncDisplayPrivate
 	char *port;
 	GdkGC *gc;
 	GdkImage *image;
+	GdkPixmap *pixmap;
 	GdkCursor *null_cursor;
 	GdkCursor *remote_cursor;
 
-#if WITH_GTKGLEXT
-	int gl_enabled;
-	GdkGLConfig *gl_config;
-	GdkGLDrawable *gl_drawable;
-	GdkGLContext *gl_context;
-	uint8_t *gl_tex_data;
-	int gl_texture_width;
-	int gl_texture_height;
-	int gl_width;
-	int gl_height;
-	GLuint gl_tex;
-#endif
-
 	struct gvnc_framebuffer fb;
 	struct coroutine coroutine;
 	struct gvnc *gvnc;
@@ -82,6 +86,7 @@ struct _VncDisplayPrivate
 	gboolean allow_lossy;
 	gboolean allow_scaling;
 	gboolean shared_flag;
+	gboolean force_size;
 
 	GSList *preferable_auths;
 };
@@ -123,7 +128,8 @@ enum
   PROP_NAME,
   PROP_LOSSY_ENCODING,
   PROP_SCALING,
-  PROP_SHARED_FLAG
+  PROP_SHARED_FLAG,
+  PROP_FORCE_SIZE
 };
 
 /* Signals */
@@ -203,6 +209,9 @@ vnc_display_get_property (GObject    *object,
       case PROP_SHARED_FLAG:
         g_value_set_boolean (value, vnc->priv->shared_flag);
 	break;
+      case PROP_FORCE_SIZE:
+        g_value_set_boolean (value, vnc->priv->force_size);
+	break;
       default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 	break;			
@@ -240,6 +249,9 @@ vnc_display_set_property (GObject      *object,
       case PROP_SHARED_FLAG:
         vnc_display_set_shared_flag (vnc, g_value_get_boolean (value));
         break;
+      case PROP_FORCE_SIZE:
+        vnc_display_set_force_size (vnc, g_value_get_boolean (value));
+        break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
         break;			
@@ -248,6 +260,7 @@ vnc_display_set_property (GObject      *object,
 
 GtkWidget *vnc_display_new(void)
 {
+	winsock_startup ();
 	return GTK_WIDGET(g_object_new(VNC_TYPE_DISPLAY, NULL));
 }
 
@@ -268,14 +281,19 @@ static GdkCursor *create_null_cursor(void)
 	return cursor;
 }
 
-static gboolean expose_event(GtkWidget *widget, GdkEventExpose *expose,
-			     gpointer data G_GNUC_UNUSED)
+static gboolean expose_event(GtkWidget *widget, GdkEventExpose *expose)
 {
 	VncDisplay *obj = VNC_DISPLAY(widget);
 	VncDisplayPrivate *priv = obj->priv;
+	int ww, wh;
+	int mx = 0, my = 0;
+#if WITH_GTK_CAIRO
+	cairo_t *cr;
+#else
 	int x, y, w, h;
 	GdkRectangle drawn;
 	GdkRegion *clear, *copy;
+#endif
 
 	GVNC_DEBUG("Expose %dx%d @ %d,%d\n",
 		   expose->area.x,
@@ -283,122 +301,96 @@ static gboolean expose_event(GtkWidget *widget, GdkEventExpose *expose,
 		   expose->area.width,
 		   expose->area.height);
 
-	if (priv->image == NULL) {
-#if WITH_GTKGLEXT
-		if (priv->gl_tex_data == NULL)
-#endif
-		{
-			GdkGC *gc = gdk_gc_new(widget->window);
-			gdk_draw_rectangle(widget->window, gc, TRUE,
-					   expose->area.x, expose->area.y,
-					   expose->area.width,
-					   expose->area.height);
-			g_object_unref(gc);
-			return TRUE;
+	gdk_drawable_get_size(widget->window, &ww, &wh);
+
+	if (ww > priv->fb.width)
+		mx = (ww - priv->fb.width) / 2;
+	if (wh > priv->fb.height)
+		my = (wh - priv->fb.height) / 2;
+
+#if WITH_GTK_CAIRO
+	cr = gdk_cairo_create(GTK_WIDGET(obj)->window);
+	cairo_rectangle(cr,
+			expose->area.x,
+			expose->area.y,
+			expose->area.width + 1,
+			expose->area.height + 1);
+	cairo_clip(cr);
+
+	/* If we don't have a pixmap, or we're not scaling, then
+	   we need to fill with background color */
+	if (!priv->pixmap ||
+	    !priv->allow_scaling) {
+		cairo_rectangle(cr, 0, 0, ww, wh);
+		/* Optionally cut out the inner area where the pixmap
+		   will be drawn. This avoids 'flashing' since we're
+		   not double-buffering. Note we're using the undocumented
+		   behaviour of drawing the rectangle from right to left
+		   to cut out the whole */
+		if (priv->pixmap)
+			cairo_rectangle(cr, mx + priv->fb.width, my,
+					-1 * priv->fb.width, priv->fb.height);
+		cairo_fill(cr);
+	}
+
+	/* Draw the VNC display */
+	if (priv->pixmap) {
+		if (priv->allow_scaling) {
+			double sx, sy;
+			/* Scale to fill window */
+			sx = (double)ww / (double)priv->fb.width;
+			sy = (double)wh / (double)priv->fb.height;
+			cairo_scale(cr, sx, sy);
+			gdk_cairo_set_source_pixmap(cr,
+						    priv->pixmap,
+						    0, 0);
+		} else {
+			gdk_cairo_set_source_pixmap(cr,
+						    priv->pixmap,
+						    mx, my);
 		}
+		cairo_paint(cr);
 	}
 
-#if WITH_GTKGLEXT
-	if (priv->gl_enabled) {
-		float rx, ry;
-		int wx = 0, wy = 0;
-		int ww = priv->gl_width, wh = priv->gl_height;
-		double scale_x, scale_y;
+	cairo_destroy(cr);
+#else
+	x = MIN(expose->area.x - mx, priv->fb.width);
+	y = MIN(expose->area.y - my, priv->fb.height);
+	w = MIN(expose->area.x + expose->area.width - mx, priv->fb.width);
+	h = MIN(expose->area.y + expose->area.height - my, priv->fb.height);
+	x = MAX(0, x);
+	y = MAX(0, y);
+	w = MAX(0, w);
+	h = MAX(0, h);
 
-		scale_x = (double)priv->gl_width / priv->fb.width;
-		scale_y = (double)priv->gl_height / priv->fb.height;
+	w -= x;
+	h -= y;
 
-		x = expose->area.x / scale_x;
-		y = expose->area.y / scale_y;
-		w = expose->area.width / scale_x;
-		h = expose->area.height / scale_y;
+	drawn.x = x + mx;
+	drawn.y = y + my;
+	drawn.width = w;
+	drawn.height = h;
 
-		y -= 5;
-		h += 10;
-		if (y < 0)
-			y = 0;
-
-		x -= 5;
-		w += 10;
-		if (x < 0)
-			x = 0;
-
-		x = MIN(x, priv->fb.width);
-		y = MIN(y, priv->fb.height);
-		w = MIN(x + w, priv->fb.width);
-		h = MIN(y + h, priv->fb.height);
-		w -= x;
-		h -= y;
-
-		gdk_gl_drawable_gl_begin(priv->gl_drawable, priv->gl_context);
-		glBindTexture(GL_TEXTURE_2D, priv->gl_tex);
-		glPixelStorei(GL_UNPACK_ROW_LENGTH, priv->fb.width);
-		glTexSubImage2D(GL_TEXTURE_2D, 0,
-				x, y, w, h,
-				GL_BGRA_EXT,
-				GL_UNSIGNED_BYTE,
-				priv->gl_tex_data +
-				y * 4 * priv->fb.width +
-				x * 4);
-		rx = (float)priv->fb.width  / priv->gl_texture_width;
-		ry = (float)priv->fb.height / priv->gl_texture_height;
-		
-		glEnable(GL_TEXTURE_2D);
-		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
-		glBegin(GL_QUADS);
-		glTexCoord2f(0,ry);  glVertex3f(wx, wy, 0);
-		glTexCoord2f(0,0);  glVertex3f(wx, wy+wh, 0);
-		glTexCoord2f(rx,0);  glVertex3f(wx+ww, wy+wh, 0);
-		glTexCoord2f(rx,ry);  glVertex3f(wx+ww, wy, 0);
-		glEnd();		
-		glDisable(GL_TEXTURE_2D);
-		glFlush();
-		gdk_gl_drawable_gl_end(priv->gl_drawable);
-	} else
-#endif
-	{
-		int mx = 0, my = 0;
-		int ww, wh;
-
-		gdk_drawable_get_size(widget->window, &ww, &wh);
-		if (ww > priv->fb.width)
-			mx = (ww - priv->fb.width) / 2;
-		if (wh > priv->fb.height)
-			my = (wh - priv->fb.height) / 2;
-
-		x = MIN(expose->area.x - mx, priv->fb.width);
-		y = MIN(expose->area.y - my, priv->fb.height);
-		w = MIN(expose->area.x + expose->area.width - mx, priv->fb.width);
-		h = MIN(expose->area.y + expose->area.height - my, priv->fb.height);
-		x = MAX(0, x);
-		y = MAX(0, y);
-		w = MAX(0, w);
-		h = MAX(0, h);
-
-		w -= x;
-		h -= y;
-
-		drawn.x = x + mx;
-		drawn.y = y + my;
-		drawn.width = w;
-		drawn.height = h;
-
-		clear = gdk_region_rectangle(&expose->area);
-		copy = gdk_region_rectangle(&drawn);
-		gdk_region_subtract(clear, copy);
+	clear = gdk_region_rectangle(&expose->area);
+	copy = gdk_region_rectangle(&drawn);
+	gdk_region_subtract(clear, copy);
 
+	if (priv->pixmap != NULL) {
 		gdk_gc_set_clip_region(priv->gc, copy);
-		gdk_draw_image(widget->window, priv->gc, priv->image,
-			       x, y, x + mx, y + my, w, h);
-
-		gdk_gc_set_clip_region(priv->gc, clear);
-		gdk_draw_rectangle(widget->window, priv->gc, TRUE, expose->area.x, expose->area.y,
-				   expose->area.width, expose->area.height);
-
-		gdk_region_destroy(clear);
-		gdk_region_destroy(copy);
+		gdk_draw_drawable(widget->window, priv->gc, priv->pixmap,
+				  x, y, x + mx, y + my, w, h);
 	}
 
+	gdk_gc_set_clip_region(priv->gc, clear);
+	gdk_draw_rectangle(widget->window, priv->gc, TRUE,
+			   expose->area.x, expose->area.y,
+			   expose->area.width, expose->area.height);
+
+	gdk_region_destroy(clear);
+	gdk_region_destroy(copy);
+#endif
+
+
 	return TRUE;
 }
 
@@ -488,8 +480,7 @@ void vnc_display_force_grab(VncDisplay *obj, gboolean enable)
 		do_pointer_ungrab(obj, FALSE);
 }
 
-static gboolean button_event(GtkWidget *widget, GdkEventButton *button,
-			     gpointer data G_GNUC_UNUSED)
+static gboolean button_event(GtkWidget *widget, GdkEventButton *button)
 {
 	VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 	int n;
@@ -520,11 +511,10 @@ static gboolean button_event(GtkWidget *widget, GdkEventButton *button,
 				   0x7FFF, 0x7FFF);
 	}
 
-	return FALSE;
+	return TRUE;
 }
 
-static gboolean scroll_event(GtkWidget *widget, GdkEventScroll *scroll,
-			     gpointer data G_GNUC_UNUSED)
+static gboolean scroll_event(GtkWidget *widget, GdkEventScroll *scroll)
 {
 	VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 	int mask;
@@ -558,14 +548,14 @@ static gboolean scroll_event(GtkWidget *widget, GdkEventScroll *scroll,
 				   0x7FFF, 0x7FFF);
 	}
 
-	return FALSE;
+	return TRUE;
 }
 
-static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion,
-			     gpointer data G_GNUC_UNUSED)
+static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
 {
 	VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 	int dx, dy;
+	int ww, wh;
 
 	if (priv->gvnc == NULL || !gvnc_is_initialized(priv->gvnc))
 		return FALSE;
@@ -576,19 +566,18 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion,
 	if (priv->read_only)
 		return FALSE;
 
-#if WITH_GTKGLEXT
-	if (priv->gl_enabled) {
-		motion->x *= priv->fb.width;
-		motion->x /= priv->gl_width;
-		motion->y *= priv->fb.height;
-		motion->y /= priv->gl_height;
-	} else
-#endif
-	{
-		int ww, wh;
+	gdk_drawable_get_size(widget->window, &ww, &wh);
+
+	if (priv->allow_scaling) {
+		double sx, sy;
+		sx = (double)priv->fb.width / (double)ww;
+		sy = (double)priv->fb.height / (double)wh;
+
+		motion->x *= sx;
+		motion->y *= sy;
+	} else {
 		int mw = 0, mh = 0;
 
-		gdk_drawable_get_size(widget->window, &ww, &wh);
 		if (ww > priv->fb.width)
 			mw = (ww - priv->fb.width) / 2;
 		if (wh > priv->fb.height)
@@ -637,16 +626,13 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion,
 	priv->last_x = (int)motion->x;
 	priv->last_y = (int)motion->y;
 
-	return FALSE;
+	return TRUE;
 }
 
-static gboolean key_event(GtkWidget *widget, GdkEventKey *key,
-			  gpointer data G_GNUC_UNUSED)
+static gboolean key_event(GtkWidget *widget, GdkEventKey *key)
 {
 	VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
-	guint keyval;
-	gint group, level;
-	GdkModifierType consumed;
+	int i;
 
 	if (priv->gvnc == NULL || !gvnc_is_initialized(priv->gvnc))
 		return FALSE;
@@ -654,35 +640,12 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key,
 	if (priv->read_only)
 		return FALSE;
 
-	/*
-	 * Key handling in VNC is screwy. The event.keyval from GTK is
-	 * interpreted relative to modifier state. This really messes
-	 * up with VNC which has no concept of modifiers - it just sees
-	 * key up & down events - the remote end interprets modifiers
-	 * itself. So if we interpret at the client end you can end up
-	 * with 'Alt' key press generating Alt_L, and key release generating
-	 * ISO_Prev_Group. This really really confuses the VNC server
-	 * with 'Alt' getting stuck on.
-	 *
-	 * So we have to redo GTK's  keycode -> keyval translation
-	 * using only the SHIFT modifier which the RFB explicitly
-	 * requires to be interpreted at client end.
-	 *
-	 * Arggggh.
-	 */
-	gdk_keymap_translate_keyboard_state(gdk_keymap_get_default(),
-					    key->hardware_keycode,
-					    key->state & (GDK_SHIFT_MASK | GDK_LOCK_MASK),
-					    key->group,
-					    &keyval,
-					    &group,
-					    &level,
-					    &consumed);
-
-	keyval = x_keymap_get_keyval_from_keycode(key->hardware_keycode, keyval);
+	GVNC_DEBUG("%s keycode: %d  state: %d  group %d, keyval: %d\n",
+		   key->type == GDK_KEY_PRESS ? "press" : "release",
+		   key->hardware_keycode, key->state, key->group, key->keyval);
 
 	/*
-	 * More VNC suckiness with key state & modifiers in particular
+	 * Some VNC suckiness with key state & modifiers in particular
 	 *
 	 * Because VNC has no concept of modifiers, we have to track what keys are
 	 * pressed and when the widget looses focus send fake key up events for all
@@ -698,67 +661,67 @@ static gboolean key_event(GtkWidget *widget, GdkEventKey *key,
 	 *
 	 * Arggggh.
 	 */
+
+	/*
+	 * First the key release handling. This is *always* run, even for Key press
+	 * events, because GTK will often merge sequential press+release pairs of
+	 * the same key into a sequence of press+press+press+press+release. VNC
+	 * servers don't like this, so we have to see if we're already pressed
+	 * send release events. So, we run the release handling code all the time.
+	 */
+	for (i = 0 ; i < (int)(sizeof(priv->down_keyval)/sizeof(priv->down_keyval[0])) ; i++) {
+		/* We were pressed, and now we're released, so... */
+		if (priv->down_scancode[i] == key->hardware_keycode) {
+			/*
+			 * ..send the key release event we're dealing with
+			 *
+			 * NB, we use priv->down_keyval[i], and not our
+			 * current 'keyval', because we need to make sure
+			 * that the release keyval is identical to the
+			 * press keyval. In some layouts, this isn't always
+			 * true, with "Tab" generating Tab on press, and
+			 * ISO_Prev_Group on release.
+			 */
+			gvnc_key_event(priv->gvnc, 0, priv->down_keyval[i], key->hardware_keycode);
+			priv->down_keyval[i] = 0;
+			priv->down_scancode[i] = 0;
+			break;
+		}
+	}
+
 	if (key->type == GDK_KEY_PRESS) {
-		int i;
 		for (i = 0 ; i < (int)(sizeof(priv->down_keyval)/sizeof(priv->down_keyval[0])) ; i++) {
 			if (priv->down_scancode[i] == 0) {
-				priv->down_keyval[i] = keyval;
+				priv->down_keyval[i] = key->keyval;
 				priv->down_scancode[i] = key->hardware_keycode;
 				/* Send the actual key event we're dealing with */
-				gvnc_key_event(priv->gvnc, 1, keyval, key->hardware_keycode);
-				break;
-			} else if (priv->down_scancode[i] == key->hardware_keycode) {
-				/* Got an press when we're already pressed ! Why ... ?
-				 *
-				 * Well, GTK merges sequential press+release pairs of the same
-				 * key so instead of press+release,press+release,press+release
-				 * we only get press+press+press+press+press+release. This
-				 * really annoys some VNC servers, so we have to un-merge
-				 * them into a sensible stream of press+release pairs
-				 */
-				/* Fake an up event for the previous down event */
-				gvnc_key_event(priv->gvnc, 0, keyval, key->hardware_keycode);
-				/* Now send our actual ldown event */
-				gvnc_key_event(priv->gvnc, 1, keyval, key->hardware_keycode);
-				break;
-			}
-		}
-	} else {
-		int i;
-		for (i = 0 ; i < (int)(sizeof(priv->down_keyval)/sizeof(priv->down_keyval[0])) ; i++) {
-			/* We were pressed, and now we're released, so... */
-			if (priv->down_scancode[i] == key->hardware_keycode) {
-				priv->down_keyval[i] = 0;
-				priv->down_scancode[i] = 0;
-				/* ..send the key release event we're dealing with */
-				gvnc_key_event(priv->gvnc, 0, keyval, key->hardware_keycode);
+				gvnc_key_event(priv->gvnc, 1, key->keyval, key->hardware_keycode);
 				break;
 			}
 		}
 	}
 
 	if (key->type == GDK_KEY_PRESS &&
-	    ((keyval == GDK_Control_L && (key->state & GDK_MOD1_MASK)) ||
-	     (keyval == GDK_Alt_L && (key->state & GDK_CONTROL_MASK)))) {
+	    ((key->keyval == GDK_Control_L && (key->state & GDK_MOD1_MASK)) ||
+	     (key->keyval == GDK_Alt_L && (key->state & GDK_CONTROL_MASK)))) {
 		if (priv->in_pointer_grab)
 			do_pointer_ungrab(VNC_DISPLAY(widget), FALSE);
-		else
+		else if (!priv->grab_keyboard || !priv->absolute)
 			do_pointer_grab(VNC_DISPLAY(widget), FALSE);
 	}
 
-	return FALSE;
+	return TRUE;
 }
 
-static gboolean enter_event(GtkWidget *widget, GdkEventCrossing *crossing,
-                            gpointer data G_GNUC_UNUSED)
+static gboolean enter_event(GtkWidget *widget, GdkEventCrossing *crossing)
 {
         VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 
         if (priv->gvnc == NULL || !gvnc_is_initialized(priv->gvnc))
-                return TRUE;
+                return FALSE;
 
         if (crossing->mode != GDK_CROSSING_NORMAL)
-                return TRUE;
+                return FALSE;
 
         if (priv->grab_keyboard)
                 do_keyboard_grab(VNC_DISPLAY(widget), FALSE);
@@ -766,32 +729,33 @@ static gboolean enter_event(GtkWidget *widget, GdkEventCrossing *crossing,
         return TRUE;
 }
 
-static gboolean leave_event(GtkWidget *widget, GdkEventCrossing *crossing,
-                            gpointer data G_GNUC_UNUSED)
+static gboolean leave_event(GtkWidget *widget, GdkEventCrossing *crossing)
 {
         VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 
         if (priv->gvnc == NULL || !gvnc_is_initialized(priv->gvnc))
-                return TRUE;
+                return FALSE;
 
         if (crossing->mode != GDK_CROSSING_NORMAL)
-                return TRUE;
+                return FALSE;
 
         if (priv->grab_keyboard)
                 do_keyboard_ungrab(VNC_DISPLAY(widget), FALSE);
 
+        if (priv->grab_pointer)
+                do_pointer_ungrab(VNC_DISPLAY(widget), FALSE);
+
         return TRUE;
 }
 
 
-static gboolean focus_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_UNUSED,
-                            gpointer data G_GNUC_UNUSED)
+static gboolean focus_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_UNUSED)
 {
         VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
 	int i;
 
         if (priv->gvnc == NULL || !gvnc_is_initialized(priv->gvnc))
-                return TRUE;
+                return FALSE;
 
 	for (i = 0 ; i < (int)(sizeof(priv->down_keyval)/sizeof(priv->down_keyval[0])) ; i++) {
 		/* We are currently pressed so... */
@@ -807,43 +771,37 @@ static gboolean focus_event(GtkWidget *widget, GdkEventFocus *focus G_GNUC_UNUSE
         return TRUE;
 }
 
-#if WITH_GTKGLEXT
-static void realize_event(GtkWidget *widget, gpointer data G_GNUC_UNUSED)
-{
-	VncDisplayPrivate *priv = VNC_DISPLAY(widget)->priv;
-
-	if (priv->gl_config == NULL)
-		return;
-
-	priv->gl_drawable = gtk_widget_get_gl_drawable(widget);
-	priv->gl_context = gtk_widget_get_gl_context(widget);
-}
-#endif
-
 static gboolean on_update(void *opaque, int x, int y, int w, int h)
 {
 	GtkWidget *widget = GTK_WIDGET(opaque);
 	VncDisplay *obj = VNC_DISPLAY(widget);
 	VncDisplayPrivate *priv = obj->priv;
+	int ww, wh;
+	GdkRectangle r = { x, y, w, h };
 
-#if WITH_GTKGLEXT
-	if (priv->gl_enabled) {
-		double scale_x, scale_y;
+	/* Copy pixbuf to pixmap */
+	gdk_gc_set_clip_rectangle(priv->gc, &r);
+	gdk_draw_image(priv->pixmap, priv->gc, priv->image,
+		       x, y, x, y, w, h);
 
-		scale_x = (double)priv->gl_width / priv->fb.width;
-		scale_y = (double)priv->gl_height / priv->fb.height;
+	gdk_drawable_get_size(widget->window, &ww, &wh);
 
-		x *= scale_x;
-		y *= scale_y;
-		w *= scale_x;
-		h *= scale_y;
-	} else
-#endif
-	{
-		int ww, wh;
+	if (priv->allow_scaling) {
+		double sx, sy;
+
+		/* Scale the VNC region to produce expose region */
+
+		sx = (double)ww / (double)priv->fb.width;
+		sy = (double)wh / (double)priv->fb.height;
+		x *= sx;
+		y *= sy;
+		w *= sx;
+		h *= sy;
+	} else {
 		int mw = 0, mh = 0;
 
-		gdk_drawable_get_size(widget->window, &ww, &wh);
+		/* Offset the VNC region to produce expose region */
+
 		if (ww > priv->fb.width)
 			mw = (ww - priv->fb.width) / 2;
 		if (wh > priv->fb.height)
@@ -853,7 +811,7 @@ static gboolean on_update(void *opaque, int x, int y, int w, int h)
 		y += mh;
 	}
 
-	gtk_widget_queue_draw_area(widget, x, y, w, h);
+	gtk_widget_queue_draw_area(widget, x, y, w + 1, h + 1);
 
 	return TRUE;
 }
@@ -866,6 +824,8 @@ static void setup_gdk_image(VncDisplay *obj, gint width, gint height)
 	visual = gdk_drawable_get_visual(GTK_WIDGET(obj)->window);
 
 	priv->image = gdk_image_new(GDK_IMAGE_FASTEST, visual, width, height);
+	priv->pixmap = gdk_pixmap_new(GTK_WIDGET(obj)->window, width, height, -1);
+
 	GVNC_DEBUG("Visual mask: %3d %3d %3d\n      shift: %3d %3d %3d\n",
 		   visual->red_mask,
 		   visual->green_mask,
@@ -888,58 +848,10 @@ static void setup_gdk_image(VncDisplay *obj, gint width, gint height)
 	priv->fb.data = (uint8_t *)priv->image->mem;
 	priv->fb.byte_order = priv->image->byte_order == GDK_LSB_FIRST ? G_LITTLE_ENDIAN : G_BIG_ENDIAN;
 
-	gtk_widget_set_size_request(GTK_WIDGET(obj), width, height);
+	if (priv->force_size)
+		gtk_widget_set_size_request(GTK_WIDGET(obj), width, height);
 }
 
-#if WITH_GTKGLEXT
-static int pow_of_2(int value)
-{
-	int i;
-	for (i = 0; value >= (1 << i); i++);
-	return (1 << i);
-}
-
-static void setup_gl_image(VncDisplay *obj, gint width, gint height)
-{
-	VncDisplayPrivate *priv = VNC_DISPLAY(obj)->priv;
-	void *dummy;
-
-	priv->gl_texture_width = pow_of_2(width);
-	priv->gl_texture_height = pow_of_2(height);
-
-	gdk_gl_drawable_gl_begin(priv->gl_drawable, priv->gl_context);
-
-	glGenTextures(1, &priv->gl_tex);
-	glBindTexture(GL_TEXTURE_2D, priv->gl_tex);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
-	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
-
-	dummy = g_malloc(priv->gl_texture_width*priv->gl_texture_height*4);
-	memset(dummy, 0, priv->gl_texture_width*priv->gl_texture_height*4);
-	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
-		     priv->gl_texture_width, priv->gl_texture_height, 0,
-		     GL_RGB, GL_UNSIGNED_BYTE,
-		     dummy);
-	g_free(dummy);
-	
-	gdk_gl_drawable_gl_end(priv->gl_drawable);
-
-	priv->gl_tex_data = g_malloc(width * height * 4);
-
-	priv->fb.red_mask = 0xFF;
-	priv->fb.green_mask = 0xFF;
-	priv->fb.blue_mask = 0xFF;
-	priv->fb.red_shift = 16;
-	priv->fb.green_shift = 8;
-	priv->fb.blue_shift = 0;
-	priv->fb.depth = 32;
-	priv->fb.bpp = 4;
-	priv->fb.width = width;
-	priv->fb.height = height;
-	priv->fb.linesize = priv->fb.width * priv->fb.bpp;
-	priv->fb.data = (uint8_t *)priv->gl_tex_data;
-}
-#endif
 
 static gboolean emit_signal_auth_cred(gpointer opaque)
 {
@@ -1015,17 +927,10 @@ static gboolean do_resize(void *opaque, int width, int height, gboolean quiet)
 		g_object_unref(priv->image);
 		priv->image = NULL;
 	}
-
-#if WITH_GTKGLEXT
-	if (priv->gl_tex_data) {
-		gdk_gl_drawable_gl_begin(priv->gl_drawable,
-					 priv->gl_context);
-		glDeleteTextures(1, &priv->gl_tex);
-		gdk_gl_drawable_gl_end(priv->gl_drawable);
-		g_free(priv->gl_tex_data);
-		priv->gl_tex_data = NULL;
+	if (priv->pixmap) {
+		g_object_unref(priv->pixmap);
+		priv->pixmap = NULL;
 	}
-#endif
 
 	if (priv->gc == NULL) {
 		priv->null_cursor = create_null_cursor();
@@ -1036,12 +941,7 @@ static gboolean do_resize(void *opaque, int width, int height, gboolean quiet)
 		priv->gc = gdk_gc_new(GTK_WIDGET(obj)->window);
 	}
 
-#if WITH_GTKGLEXT
-	if (priv->gl_enabled)
-		setup_gl_image(obj, width, height);
-	else
-#endif
-		setup_gdk_image(obj, width, height);
+	setup_gdk_image(obj, width, height);
 
 	gvnc_set_local(priv->gvnc, &priv->fb);
 
@@ -1068,156 +968,28 @@ static gboolean on_pixel_format(void *opaque,
         return do_resize(opaque, priv->fb.width, priv->fb.height, TRUE);
 }
 
-#if WITH_GTKGLEXT
-static void build_gl_image_from_gdk(uint32_t *data, GdkImage *image)
+static gboolean on_get_preferred_pixel_format(void *opaque,
+	struct gvnc_pixel_format *fmt)
 {
-	GdkVisual *visual;
-	int i, j;
-	uint8_t *row;
+	VncDisplay *obj = VNC_DISPLAY(opaque);
+	GdkVisual *v =  gdk_drawable_get_visual(GTK_WIDGET(obj)->window);
 
-	visual = image->visual;
-	row = image->mem;
-	for (j = 0; j < image->height; j++) {
-		uint8_t *src = row;
-		for (i = 0; i < image->width; i++) {
-			uint32_t pixel = 0;
-			switch (image->bpp) {
-			case 1:
-				pixel = *(uint8_t *)src;
-				break;
-			case 2:
-				pixel = *(uint16_t *)src;
-				break;
-			case 4:
-				pixel = *(uint32_t *)src;
-				break;
-			}
-			*data = ((pixel & visual->red_mask) >> visual->red_shift) << (24 - visual->red_prec) |
-				((pixel & visual->green_mask) >> visual->green_shift) << (16 - visual->green_prec) |
-				((pixel & visual->blue_mask) >> visual->blue_shift) << (8 - visual->blue_prec);
-			src += image->bpp;
-			data++;
-		}
-		row += image->bpl;
+	GVNC_DEBUG("Setting pixel format to true color\n");
 
-	}
+	fmt->true_color_flag = 1;
+	fmt->depth = v->depth;
+	fmt->bits_per_pixel = v->depth > 16 ? 32 : v->depth;
+	fmt->red_max = v->red_mask >> v->red_shift;
+	fmt->green_max = v->green_mask >> v->green_shift;
+	fmt->blue_max = v->blue_mask >> v->blue_shift;
+	fmt->red_shift = v->red_shift;
+	fmt->green_shift = v->green_shift;
+	fmt->blue_shift = v->blue_shift;
+	fmt->byte_order = v->byte_order == GDK_LSB_FIRST ? G_BIG_ENDIAN : G_LITTLE_ENDIAN;
+
+	return TRUE;
 }
 
-static void build_gdk_image_from_gl(GdkImage *image, uint32_t *data)
-{
-	GdkVisual *visual;
-	int i, j;
-	uint8_t *row;
-
-	visual = image->visual;
-	row = image->mem;
-	for (j = 0; j < image->height; j++) {
-		uint8_t *dst = row;
-		for (i = 0; i < image->width; i++) {
-			uint32_t pixel;
-
-			pixel = (((*data >> (24 - visual->red_prec)) << visual->red_shift) & visual->red_mask) |
-				(((*data >> (16 - visual->green_prec)) << visual->green_shift) & visual->green_mask) |
-				(((*data >> (8 - visual->blue_prec)) << visual->blue_shift) & visual->blue_mask);
-
-			switch (image->bpp) {
-			case 1:
-				*(uint8_t *)dst = pixel;
-				break;
-			case 2:
-				*(uint16_t *)dst = pixel;
-				break;
-			case 4:
-				*(uint32_t *)dst = pixel;
-				break;
-			}
-			dst += image->bpp;
-			data++;
-		}
-		row += image->bpl;
-	}
-}
-
-static void scale_display(VncDisplay *obj, gint width, gint height)
-{
-	VncDisplayPrivate *priv = VNC_DISPLAY(obj)->priv;
-
-	if (priv->gl_drawable == NULL)
-		return;
-
-	if (priv->gl_enabled == 0) {
-		GdkImage *image;
-
-		priv->gl_enabled = 1;
-
-		image = priv->image;
-		priv->image = NULL;
-	
-		do_resize(obj, priv->fb.width, priv->fb.height, TRUE);
-		build_gl_image_from_gdk((uint32_t *)priv->fb.data, image);
-
-		g_object_unref(image);
-	}
-
-	priv->gl_width = width;
-	priv->gl_height = height;
-
-	gdk_gl_drawable_gl_begin(priv->gl_drawable, priv->gl_context);
-	glClearColor (0.0, 0.0, 0.0, 0.0);
-	glShadeModel(GL_FLAT);
-	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
-    
-	glViewport(0, 0, priv->gl_width, priv->gl_height);
-	glMatrixMode(GL_PROJECTION);
-	glLoadIdentity();
-	glOrtho(0.0, priv->gl_width, 0.0, priv->gl_height, -1, 1);
-	glMatrixMode(GL_MODELVIEW);
-	glLoadIdentity();
-	gdk_gl_drawable_gl_end(priv->gl_drawable);
-}
-
-static void rescale_display(VncDisplay *obj, gint width, gint height)
-{
-	VncDisplayPrivate *priv = obj->priv;
-
-	if (priv->allow_scaling && 
-	    (priv->fb.width != width ||
-	     priv->fb.height != height))
-		scale_display(obj, width, height);
-	else if (priv->gl_enabled) {
-		void *data;
-		priv->gl_enabled = 0;
-
-		data = priv->gl_tex_data;
-		priv->gl_tex_data = NULL;
-
-		do_resize(GTK_WIDGET(obj), priv->fb.width, priv->fb.height, TRUE);
-
-		build_gdk_image_from_gl(priv->image, (uint32_t *)data);
-		gdk_gl_drawable_gl_begin(priv->gl_drawable,
-					 priv->gl_context);
-		glDeleteTextures(1, &priv->gl_tex);
-		gdk_gl_drawable_gl_end(priv->gl_drawable);
-		g_free(data);
-	}
-}
-
-static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *configure,
-				gpointer data G_GNUC_UNUSED)
-{
-	VncDisplay *obj = VNC_DISPLAY(widget);
-	VncDisplayPrivate *priv = obj->priv;
-
-	if (priv->fb.data == NULL)
-		return FALSE;
-
-	rescale_display(VNC_DISPLAY(widget),
-			configure->width, configure->height);
-	
-	return FALSE;
-}
-#endif
-
 static gboolean on_pointer_type_change(void *opaque, int absolute)
 {
 	VncDisplay *obj = VNC_DISPLAY(opaque);
@@ -1456,6 +1228,7 @@ static const struct gvnc_ops vnc_display_ops = {
 	.server_cut_text = on_server_cut_text,
 	.bell = on_bell,
 	.render_jpeg = on_render_jpeg,
+	.get_preferred_pixel_format = on_get_preferred_pixel_format
 };
 
 /* we use an idle function to allow the coroutine to exit before we actually
@@ -1470,15 +1243,15 @@ static gboolean delayed_unref_object(gpointer data)
 		g_object_unref(obj->priv->image);
 		obj->priv->image = NULL;
 	}
-
-#if WITH_GTKGLEXT
-	if (obj->priv->gl_tex_data)
-		g_free(obj->priv->gl_tex_data);
-	obj->priv->gl_tex_data = NULL;
-	obj->priv->gl_enabled = 0;
-#endif
+	if (obj->priv->pixmap) {
+		g_object_unref(obj->priv->pixmap);
+		obj->priv->pixmap = NULL;
+	}
 
 	g_object_unref(G_OBJECT(data));
+
+	winsock_cleanup ();
+
 	return FALSE;
 }
 
@@ -1652,15 +1425,6 @@ void vnc_display_close(VncDisplay *obj)
 		gvnc_shutdown(priv->gvnc);
 	}
 
-#if WITH_GTKGLEXT
-	if (priv->gl_tex_data) {
-		gdk_gl_drawable_gl_begin(priv->gl_drawable,
-					 priv->gl_context);
-		glDeleteTextures(1, &priv->gl_tex);
-		gdk_gl_drawable_gl_end(priv->gl_drawable);
-	}
-#endif
-
 	if (widget->window) {
 		gint width, height;
 
@@ -1749,24 +1513,6 @@ static void vnc_display_finalize (GObject *obj)
 	gvnc_free(priv->gvnc);
 	display->priv->gvnc = NULL;
 
-#if WITH_GTKGLEXT
-	if (priv->gl_enabled) {
-		gdk_gl_drawable_gl_begin(priv->gl_drawable,
-					 priv->gl_context);
-		glDeleteTextures(1, &priv->gl_tex);
-		gdk_gl_drawable_gl_end(priv->gl_drawable);
-		if (priv->gl_tex_data) {
-			g_free(priv->gl_tex_data);
-			priv->gl_tex_data = NULL;
-		}
-	}
-
-	if (priv->gl_config) {
-		g_object_unref(G_OBJECT(priv->gl_config));
-		priv->gl_config = NULL;
-	}
-#endif
-
 	if (priv->image) {
 		g_object_unref(priv->image);
 		priv->image = NULL;
@@ -1781,6 +1527,18 @@ static void vnc_display_class_init(VncDisplayClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 	GtkObjectClass *gtkobject_class = GTK_OBJECT_CLASS (klass);
+	GtkWidgetClass *gtkwidget_class = GTK_WIDGET_CLASS (klass);
+
+	gtkwidget_class->expose_event = expose_event;
+	gtkwidget_class->motion_notify_event = motion_event;
+	gtkwidget_class->button_press_event = button_event;
+	gtkwidget_class->button_release_event = button_event;
+	gtkwidget_class->scroll_event = scroll_event;
+	gtkwidget_class->key_press_event = key_event;
+	gtkwidget_class->key_release_event = key_event;
+	gtkwidget_class->enter_notify_event = enter_event;
+	gtkwidget_class->leave_notify_event = leave_event;
+	gtkwidget_class->focus_out_event = focus_event;
 
 	object_class->finalize = vnc_display_finalize;
 	object_class->get_property = vnc_display_get_property;
@@ -1899,6 +1657,17 @@ static void vnc_display_class_init(VncDisplayClass *klass)
 								G_PARAM_STATIC_NAME |
 								G_PARAM_STATIC_NICK |
 								G_PARAM_STATIC_BLURB));
+	g_object_class_install_property (object_class,
+					 PROP_FORCE_SIZE,
+					 g_param_spec_boolean ( "force-size",
+								"Force widget size",
+								"Whether we should define the widget size",
+								TRUE,
+								G_PARAM_READWRITE |
+								G_PARAM_CONSTRUCT |
+								G_PARAM_STATIC_NAME |
+								G_PARAM_STATIC_NICK |
+								G_PARAM_STATIC_BLURB));
 
 	signalCredParam = g_param_spec_enum("credential",
 					    "credential",
@@ -2062,33 +1831,6 @@ static void vnc_display_init(VncDisplay *display)
 	GtkWidget *widget = GTK_WIDGET(display);
 	VncDisplayPrivate *priv;
 
-	g_signal_connect(obj, "expose-event",
-			 G_CALLBACK(expose_event), NULL);
-	g_signal_connect(obj, "motion-notify-event",
-			 G_CALLBACK(motion_event), NULL);
-	g_signal_connect(obj, "button-press-event",
-			 G_CALLBACK(button_event), NULL);
-	g_signal_connect(obj, "button-release-event",
-			 G_CALLBACK(button_event), NULL);
-	g_signal_connect(obj, "scroll-event",
-			 G_CALLBACK(scroll_event), NULL);
-	g_signal_connect(obj, "key-press-event",
-			 G_CALLBACK(key_event), NULL);
-	g_signal_connect(obj, "key-release-event",
-			 G_CALLBACK(key_event), NULL);
-	g_signal_connect(obj, "enter-notify-event",
-			 G_CALLBACK(enter_event), NULL);
-	g_signal_connect(obj, "leave-notify-event",
-			 G_CALLBACK(leave_event), NULL);
-	g_signal_connect(obj, "focus-out-event",
-			 G_CALLBACK(focus_event), NULL);
-#if WITH_GTKGLEXT
-	g_signal_connect(obj, "realize",
-			 G_CALLBACK(realize_event), NULL);
-	g_signal_connect(obj, "configure-event",
-			 G_CALLBACK(configure_event), NULL);
-#endif
-
 	GTK_WIDGET_SET_FLAGS(obj, GTK_CAN_FOCUS);
 
 	gtk_widget_add_events(widget,
@@ -2115,29 +1857,13 @@ static void vnc_display_init(VncDisplay *display)
 	priv->grab_keyboard = FALSE;
 	priv->local_pointer = FALSE;
 	priv->shared_flag = FALSE;
+	priv->force_size = TRUE;
 
 	priv->preferable_auths = g_slist_append (priv->preferable_auths, GUINT_TO_POINTER (GVNC_AUTH_VENCRYPT));
 	priv->preferable_auths = g_slist_append (priv->preferable_auths, GUINT_TO_POINTER (GVNC_AUTH_TLS));
 	priv->preferable_auths = g_slist_append (priv->preferable_auths, GUINT_TO_POINTER (GVNC_AUTH_VNC));
 	priv->preferable_auths = g_slist_append (priv->preferable_auths, GUINT_TO_POINTER (GVNC_AUTH_NONE));
 
-#if WITH_GTKGLEXT
-	if (gtk_gl_init_check(NULL, NULL)) {
-		priv->gl_config = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGB |
-							    GDK_GL_MODE_DEPTH);
-		if (!gtk_widget_set_gl_capability(widget,
-						  priv->gl_config,
-						  NULL,
-						  TRUE,
-						  GDK_GL_RGBA_TYPE)) {
-			g_warning("Could not enable OpenGL");
-			g_object_unref(G_OBJECT(priv->gl_config));
-			priv->gl_config = NULL;
-		}
-	} else
-		priv->gl_config = NULL;
-#endif
-
 	priv->gvnc = gvnc_new(&vnc_display_ops, obj);
 }
 
@@ -2161,33 +1887,44 @@ static int vnc_display_best_path(char *buf,
 
 static int vnc_display_set_x509_credential(VncDisplay *obj, const char *name)
 {
-	char sysdir[PATH_MAX], userdir[PATH_MAX];
-	struct passwd *pw;
 	char file[PATH_MAX];
+	char sysdir[PATH_MAX];
+#ifndef WIN32
+	char userdir[PATH_MAX];
+	struct passwd *pw;
 	char *dirs[] = { sysdir, userdir };
+#else
+	char *dirs[] = { sysdir };
+#endif
 
 	strncpy(sysdir, SYSCONFDIR "/pki", PATH_MAX-1);
 	sysdir[PATH_MAX-1] = '\0';
 
+#ifndef WIN32
 	if (!(pw = getpwuid(getuid())))
 		return TRUE;
 
 	snprintf(userdir, PATH_MAX-1, "%s/.pki", pw->pw_dir);
 	userdir[PATH_MAX-1] = '\0';
+#endif
 
-	if (vnc_display_best_path(file, PATH_MAX, "CA", "cacert.pem", dirs, 2) < 0)
+	if (vnc_display_best_path(file, PATH_MAX, "CA", "cacert.pem",
+				  dirs, sizeof(dirs)/sizeof(dirs[0])) < 0)
 		return TRUE;
 	gvnc_set_credential_x509_cacert(obj->priv->gvnc, file);
 
 	/* Don't mind failures of CRL */
-	if (vnc_display_best_path(file, PATH_MAX, "CA", "cacrl.pem", dirs, 2) == 0)
+	if (vnc_display_best_path(file, PATH_MAX, "CA", "cacrl.pem",
+				  dirs, sizeof(dirs)/sizeof(dirs[0])) == 0)
 		gvnc_set_credential_x509_cacert(obj->priv->gvnc, file);
 
 	/* Set client key & cert if we have them. Server will reject auth
 	 * if it decides it requires them*/
-	if (vnc_display_best_path(file, PATH_MAX, name, "private/clientkey.pem", dirs, 2) == 0)
+	if (vnc_display_best_path(file, PATH_MAX, name, "private/clientkey.pem",
+				  dirs, sizeof(dirs)/sizeof(dirs[0])) == 0)
 		gvnc_set_credential_x509_key(obj->priv->gvnc, file);
-	if (vnc_display_best_path(file, PATH_MAX, name, "clientcert.pem", dirs, 2) == 0)
+	if (vnc_display_best_path(file, PATH_MAX, name, "clientcert.pem",
+				  dirs, sizeof(dirs)/sizeof(dirs[0])) == 0)
 		gvnc_set_credential_x509_cert(obj->priv->gvnc, file);
 
 	return FALSE;
@@ -2345,33 +2082,43 @@ void vnc_display_set_shared_flag(VncDisplay *obj, gboolean shared)
 	obj->priv->shared_flag = shared;
 }
 
-#if WITH_GTKGLEXT
-gboolean vnc_display_set_scaling(VncDisplay *obj, gboolean enable)
+#if WITH_GTK_CAIRO
+gboolean vnc_display_set_scaling(VncDisplay *obj,
+				 gboolean enable)
 {
-	GtkWidget *widget = GTK_WIDGET(obj);
-	gint width, height;
+	int ww, wh;
 
-	g_return_val_if_fail (VNC_IS_DISPLAY (obj), FALSE);
-	if (obj->priv->gl_config == NULL)
-		return FALSE;
-	
 	obj->priv->allow_scaling = enable;
-	if (gvnc_is_open(obj->priv->gvnc) && widget->window) {
-		gdk_drawable_get_size(widget->window, &width, &height);
-		rescale_display(obj, width, height);
-		gtk_widget_queue_draw_area(widget, 0, 0, width, height);
+
+	if (obj->priv->pixmap != NULL) {
+		gdk_drawable_get_size(GTK_WIDGET(obj)->window, &ww, &wh);
+		gtk_widget_queue_draw_area(GTK_WIDGET(obj), 0, 0, ww, wh);
 	}
 
 	return TRUE;
 }
 #else
 gboolean vnc_display_set_scaling(VncDisplay *obj G_GNUC_UNUSED,
-	gboolean enable G_GNUC_UNUSED)
+				 gboolean enable G_GNUC_UNUSED)
 {
 	return FALSE;
 }
 #endif
 
+
+void vnc_display_set_force_size(VncDisplay *obj, gboolean enabled)
+{
+	g_return_if_fail (VNC_IS_DISPLAY (obj));
+	obj->priv->force_size = enabled;
+}
+
+gboolean vnc_display_get_force_size(VncDisplay *obj)
+{
+	g_return_val_if_fail (VNC_IS_DISPLAY (obj), FALSE);
+
+	return obj->priv->force_size;
+}
+
 gboolean vnc_display_get_scaling(VncDisplay *obj)
 {
 	g_return_val_if_fail (VNC_IS_DISPLAY (obj), FALSE);
@@ -2438,6 +2185,49 @@ vnc_display_get_option_group (void)
   return group;
 }
 
+#ifdef WIN32
+
+/* On Windows, we must call WSAStartup before using any sockets and we
+ * must call WSACleanup afterwards.  And we have to balance any calls
+ * to WSAStartup with a corresponding call to WSACleanup.
+ *
+ * Note that Wine lets you do socket calls anyway, but real Windows
+ * doesn't. (http://bugs.winehq.org/show_bug.cgi?id=11965)
+ */
+
+static void
+winsock_startup (void)
+{
+	WORD winsock_version, err;
+	WSADATA winsock_data;
+
+	/* http://msdn2.microsoft.com/en-us/library/ms742213.aspx */
+	winsock_version = MAKEWORD (2, 2);
+	err = WSAStartup (winsock_version, &winsock_data);
+	if (err != 0)
+		GVNC_DEBUG ("ignored error %d from WSAStartup", err);
+}
+
+static void
+winsock_cleanup (void)
+{
+	WSACleanup ();
+}
+
+#else /* !WIN32 */
+
+static void
+winsock_startup (void)
+{
+}
+
+static void
+winsock_cleanup (void)
+{
+}
+
+#endif /* !WIN32 */
+
 /*
  * Local variables:
  *  c-indent-level: 8
